/*
 * cgi-proxy - relays a CGI GET request over a socket, and hands the answer
 * back to web server.
 * 
 * Copyright by Samuli Krkkinen <skarkkai@woods.iki.fi>.
 *
 * Released under GNU General Public License (GPL).
 *
 * TAB size 4
 *
 * If compiling under Solaris, give cc command line switches -lnsl -lsocket
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>
#include <netdb.h>
#include <errno.h>

#define CONF_FILE "/etc/wcal.conf"
/* default values, can be overriden in CONF_FILE */
#define ADDRESS	"127.0.0.1"
#define PORT	"13134"

#define NENV 1000				/* max number of environment variables */

char *address;
char *port;
char *nvmatch ();
char *newenv[NENV];
char *password;

void print_error (const char *message)
{
	if (! message)
		return;
	printf ("Content-Type: text/html\n\n<HTML><BODY><H2>cgi-proxy error: %s</H2></BODY></HTML>\n", message);
	return;
}

/* create a socket connected to port and hostname given in globals */
int create_client_socket()
{
    struct sockaddr_in sa;
    struct hostent *hp;
    int s;
    long addr;

    bzero(&sa, sizeof(sa));
    if ((addr = inet_addr(address)) != -1) {
		/* is Internet addr in octet notation */
		memcpy((char *) &sa.sin_addr, &addr, sizeof(addr)); /* set address */
		sa.sin_family = AF_INET;
    } else {
		/* do we know the host's address? */
		if ((hp = gethostbyname(address)) == NULL) {
		    return -1;
		}
		memcpy((char *) &sa.sin_addr, hp->h_addr, hp->h_length);
		sa.sin_family = hp->h_addrtype;
    }

    sa.sin_port = htons((u_short) atoi (port));

    if ((s = socket(sa.sin_family, SOCK_STREAM, 0)) < 0) { /* get socket */
		return -2;
    }
    if (connect(s, (struct sockaddr *) &sa, sizeof(sa)) < 0) {                  /* connect */
		close(s);
		return -3;
    }
    return s;
}

void doit (void)
{
	int s;
	char buf[8192];
	char curwd[1024];
	char *basename;
	int readlen;
	char **p = newenv;

	if ((s = create_client_socket()) < 0) {
		if (s == -1) {
			snprintf (buf, 8192, "gethostbyname (%s): error number %d", address, h_errno);
			print_error (buf);
		} else if (s == -2) {
			snprintf (buf, 8192, "socket: %s", sys_errlist[errno]);
			print_error (buf);
		} else if (s == -3) {
			snprintf (buf, 8192, "connect to %s: %s", address, sys_errlist[errno]);
			print_error (buf);
		} else {
			print_error ("can't happen error");
		}
		exit (1);
	}

	/*
	 * write entire environment into the socket
	 * (wcald is interested at least in QUERY_STRING and REMOTE_ADDR)
	 */
	while (*p != NULL) {
		if (strchr (*p, '\n')) {
			(void) *p++;
			continue;	/* don't send variables with newlines in them, such as Apache 1.3+ Serversignature */
		}
		write (s, *p, strlen (*p));
		write (s, "\n", 1);
		(void) *p++;
	}

	if (! getcwd (curwd, (size_t) 1024)) {
		perror ("getcwd");
		exit (1);
	}
	basename = strrchr (curwd, '/');
	basename++;

	write (s, "__password=", strlen ("__password="));
	write (s, password, strlen (password)); write (s, "\n", 1);

	write (s, "__user=", strlen ("__user="));
	write (s, basename, strlen (basename)); write (s, "\n", 1);

	sprintf (buf, "%d", getegid());
	write (s, "__gid=", strlen ("__gid="));
	write (s, buf, strlen (buf)); write (s, "\n", 1);

	write (s, "\n", 1);
	while ((readlen = read (s, buf, 8192)) != 0) {
		write (1, buf, readlen);
	}

	close (s);
	return;
}

void addname (char *arg)
{
	register char **p;

	for (p = newenv; *p != NULL; p++) {
		if (p >= &newenv[NENV - 1]) {
			(void) fputs ("too many values in environment\n",
						  stderr);
		}
		if (nvmatch (arg, *p) != NULL)
			break;
	}
	*p = arg;
	return;
}

char * nvmatch (char *s1, char *s2)
{

	while (*s1 == *s2++)
		if (*s1++ == '=')
			return (s2);
	if (*s1 == '\0' && *(s2 - 1) == '=')
		return (s2);
	return (NULL);
}

char *remove_leading_trailing_space (char *s)
{
	char *start;
	while (*s == ' ')
		s++;
	start = s;
	if (! *s)
		return start;
	while (*s)
		s++;
	s--;
	while (*s == ' ')
		s--;
	*(s+1) = '\000';
	return start;
}

int get_conf_line (FILE *file, char **key, char **value)
{
	static char buffer[10240];
	char *lkey, *lvalue;
	int i;
	
	if (! fgets (buffer, 10240, file)) {
		return (0);
	}
	for (i = 0; buffer[i] != '\000'; i++) {
		if (buffer[i] == '\t' || buffer[i] == '\n') {
			buffer[i] = ' ';
		}
	}
	lkey = buffer;
	lvalue = strchr (buffer, '=');
	if (lvalue == NULL) {
		lvalue = "";
	} else {
		*lvalue = '\000';
		lvalue++;
	}
	lkey = remove_leading_trailing_space (lkey);
	for (i = 0; lkey[i] != '\000'; i++) {
		lkey[i] = tolower (lkey[i]);
	}
	lvalue = remove_leading_trailing_space (lvalue);
	*key = lkey;
	*value = lvalue;
	return (1);
}

void read_config ()
{
	FILE *file;
	char *key, *value;
	
	file = fopen (CONF_FILE, "r");
	if (! file) {
		print_error ("can't open " CONF_FILE " for reading");
		exit (1);
	}
	address = NULL;
	port = NULL;
	password = NULL;
	while (get_conf_line (file, &key, &value)) {
		if (strcmp (key, "password") == 0) {
			password = strdup (value);
		} else if (strcmp (key, "address") == 0) {
			address = strdup (value);	/* host name/address on which wcald runs */
		} else if (strcmp (key, "port") == 0) {
			port = strdup (value);		/* port on which wcald listens */
		}
	}
	if (! password) {
		print_error ("can't find password from configuration file");
		exit (1);
	}
	if (! address) {
		address = ADDRESS;	/* default value */
	}
	if (! port) {
		port = PORT;		/* default value */
	}
	fclose (file);

	return;
}

int main (int argc, char **argv, char **envp)
{
	read_config ();

	/* Copy environment into newenv */
	for (; *envp != NULL; envp++) {
		if (strchr (*envp, '=') != *envp)
			addname (*envp);
	}

	doit ();
	return 0;
}
